#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
# Barcerlona Supercomputing Center - Centro Nacional de Supercomputacion
#        Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: Jesús Peña-Izquierdo (jesus.pena@bsc.es)
#             -
#       
import xarray
import matplotlib.pyplot as plt
import numpy as np
import os
import sys
import pandas as pd
from scipy.stats import kurtosis, skew
from statsmodels.stats.stattools import medcouple
import time
import datetime


class valid_ranges_test:

    def __init__(self,
                 path,
                 type = None,
                 nbins = 30,
                 OutIn = True,
                 intial_xbin_factor = 2
                ):
        
        self.type = type
        self.path = path
        self.log_path = '/data/DQC_LOGS'
        self.files = list_files_folder()
        self.nbins = nbins
        self.OutIn = OutIn
        self.initial_xbin_factor = initial_xbin_factor
        self.stats_varnames = ['Minimum', 'Maximum',
                    'Mean', 'Median',
                    'Skewness', 'Kurtosis',
                    'Std', 'Std_menos', 'Std_mas', 
                    'Mad', 'Mad_menos', 'Mad_mas',
                    'MC', 'Q1', 'Q3', 'IQD']
        self.stats_k_varnames = ['k_sd_min', 'k_sd_max',
                    'k_sd_menos_min', 'k_sd_mas_max',
                    'k_mad_min', 'k_mad_max',
                    'k_mad_menos_min', 'k_mad_mas_max']
    
    
    def list_files_folder(self):
        list_files = []
        
        if os.path.exists(self.path) and os.path.isdir(self.path):
            for root, dirs, files in os.walk(self.path):
                    for f in files:
                        if self.type = 'netcdf' and f.endswith("nc"):
                            list_files.append(os.path.join(root,f))
                        elif self.type = 'grib' and f.endswith("grib"):
                            list_files.append(os.path.join(root,f))
            
            return sorted(list_files)
        else:
            raise Exception('\n\tGiven path is not a valid path\n')
        
    def initiate_logging(self):
        # Initiate the logging
        time0 = self.timefortitle
        path_logs = self.log_path

        if not os.path.exists(path_logs):
            os.makedirs(path_logs)

        default_log_name = '_'.join(
            [
                'LOG_TEST06_VALID_RANGE',
                (self.type).upper(),
                time0,
                '.log'
            ]
        )

        default_path = os.path.join(
            path_logs,
            default_log_name
        )
        
        logging.basicConfig(
            filename=default_path,
            format='%(levelname)s:%(asctime)s:%(message)s', 
            level=logging.INFO
        )

        stringA = '\n' + '#'*80 + '\n' + '#'*80
        stringA += '\n RUNNING STANDARD COMPLIANCE TEST FOR FOLDER:'
        stringA += '\n ' + self.path
        stringA += '\n AGAINST ' + (self.type).upper() + ' CONVENTION '
        stringA += '\n EXECUTION TIME: ' + self.timeforinfo
        stringA += '\n' + '#'*80

        logging.info(stringA)

     
    #Loading data depending on fileformat
    def load_datafile(self,
                      filepath
                     ):
        if filepath.name.endswith('nc'):
            data = xarray.open_dataset(filepath).to_array().values
            data_shape = data.shape
                    
        if filepath.name.endswith('grib'):
            print('WILL BE READY SOON')
            
        data = data[~np.isnan(data)].flatten()
            
        return data, data_shape
    
    #Producing a list of statistical estimators
    def Stats(self,
              data
             ):
        #data to be a flatten numpy array

        Mean = np.mean(data)
        Median = np.median(data)
        Std = np.std(data)

        #MC = iter_medcouple(data, ni = 1000)
        MC = skew(data)
        Q1 = np.quantile(data,0.25)
        Q3 = np.quantile(data,0.75)
        IQD = Q3 - Q1

        #Mad
        diff = np.abs(data - Median)
        Mad = np.median(diff)

        diff_menos = np.abs(data[data <= Median] - Median)
        #Mad_menos = np.median(diff_menos)
        Mad_menos = np.mean(diff_menos)

        diff_mas = np.abs(data[data >= Median] - Median)
        #Mad_mas = np.median(diff_mas)
        Mad_mas = np.mean(diff_mas)

        Std_menos =  np.sqrt( np.mean( np.square(data[data<=np.mean(data)] - np.mean(data)))) #np.mean( np.abs(data[data<=np.mean(data)] - np.mean(data)))
        Std_mas = np.sqrt( np.mean( np.square(data[data>=np.mean(data)] - np.mean(data)))) #np.mean( np.abs(data[data>=np.mean(data)] - np.mean(data)))

        Kurt = kurtosis(data)
        Skew = skew(data)

        stats_list = [Mean, Median, Kurt, Skew, Std, Std_menos, Std_mas, Mad, Mad_menos, Mad_mas, MC, Q1, Q3, IQD]

    return stats_list


    # Monitoring the absolute minimum and maximun values during online computation
    def minmax_monitor(self,
                       data,
                       minmax_record
                      ):

        if len(minmax_record)==2:
            min_record, max_record = minmax_record

            min_record = min([np.min(data), min_record])
            max_record = max([np.max(data), max_record])

        else:
            min_record = np.min(data)
            max_record = np.max(data)    

        minmax_record = np.array([min_record, max_record])
        return minmax_record

    # Weighted average for online computation 
    def weight_avg(self,
                   N__Ni,
                   Var_n__Var_i
                  ):

        N, Ni = N__Ni
        Var_n, Var_i = Var_n__Var_i

        Var_M = []

        for var_n_k, var_i_k in zip(Var_n, Var_i):
            var_m_k = (var_n_k*N + var_i_k*Ni) / (N + Ni)
            Var_M.append(var_m_k)

        return Var_M

    # Computing data distribution online
    def online_distribution(self,
                            data,
                            limits_record,
                            distribution_record
                            ):

        dlimits = np.diff(limits_record)
        limits_record_ext = [limits_record[0] - dlimits / initial_xbin_factor, limits_record[1] + dlimits / initial_xbin_factor]
        n_i, xbin = np.histogram(data, bins = bins, range = limits_record_ext)
        #n_i = n_i / normalization_factor

        #For accumulating outside xbin values within the edges
        if OutIn:
            n_outmin = len(np.where(data < limits_record_ext[0])[0])
            n_outmax = len(np.where(data > limits_record_ext[1])[0])
            n_i[0] = n_i[0] + n_outmax
            n_i[-1] = n_i[-1] + n_outmin

        distribution_record = distribution_record + n_i

        return distribution_record, xbin

    # Averaging statistical estimators online
    def avg_online_stats(self,
                         bins = 50,
                         OutIn = True):

        iMean = 0
        iMed = 1
        iStd = 4
        iMad = 7

        list_files = self.list_files
        
        Nt = 0
        minmax_record = []

        for ifile in list_files:

            #Loading single files
            data, data_shape = load_datafile(self, ifile)
            Nt_i,Ny,Nx = data_shape
            
            #MinMax record
            minmax_record =  minmax_monitor(data, minmax_record)

            #Stats
            stats_data_i = Stats(data)
            if Nt == 0:
                stats_data_n = [0,] * len(stats_data_i)
                limits_record = minmax_record
                distribution_record = np.zeros((bins,))

            stats_data_n = weight_avg( (Nt, Nt_i), (stats_data_n, stats_data_i))

            distribution_record, xbin = online_distribution(data, limits_record, distribution_record, 
                                                            bins = bins, OutIn = OutIn)

        #Proportion between Mean, sd and min/max
        k_sd = np.empty((3,2))

        #Proportion between Mads and (Min/Max - Median)
        k_mad = np.empty((3,2))
        for i in range(3):
            if stats_data_n[iMad + i] == 0:
                k_mad[i,:] = np.zeros((2,))
            else:
                k_mad[i,:] = np.abs(minmax_record - stats_data_n[iMed]) / stats_data_n[iMad + i]
            k_sd[i,:] = np.abs(minmax_record - stats_data_n[iMean]) / stats_data_n[iStd + i]

        k_sd[1,1] = k_sd[2,1] 
        k_sd = k_sd[:2,:].flatten()

        k_mad[1,1] = k_mad[2,1] #First row (Min/Max - Median), second row (Min/Max - Median_menos/Median_mas)
        k_mad = k_mad[:2,:].flatten()        


        Nt += Nt_i

        stats_var = np.concatenate((minmax_record, stats_data_n))
        stats_k =np.concatenate((k_sd, k_mad))

        logging.info('{} analysed'.format(ifile))
            
           
        return stats_var, stats_k, distribution_record, xbin
    
    
    def check(self,
              k_sd_REF = 10
             ):
        
        result_base_filename = 'TEST06_VALID_RANGE_RESULTS'
        islash = [i for i, ltr in enumerate(s) if ltr == '/'][-1]
        varname = self.path[islash+1:]
        
        
        stats_var, stats_k, _, _ = avg_online_stats(bins = 50, OutIn = True)
                      
        result_filename = os.path.join(self.log_path,
                                        '_'.join(result_base_filename, varname))
        
        
        k_sd = stats_k[0:2]
        if k_sd.any() > k_sd_REF:
            result = 'Valid range test NOT PASSED: [k_sd_min, k_sd_max] = {} > {}'.format(str(k_sd), str(k_sd,REF))
        else:
            result = 'Valid range test PASSED: [k_sd_min, k_sd_max] = {} < {}'.format(str(k_sd), str(k_sd,REF))
        
        logging.info('Writting results to {}'.format(result_filename))    
        with open(result_filename, "w") as rf:
            json.dump(result, rf)
            json.dump(self.stats_varnames, rf, indent = 4)
            json.dump(self.stats_var, rf, indent = 4)
            json.dump(self.stats_k_varnames, rf, indent = 4)
            json.dump(self.stats_k, rf, indent = 4)




#     def monitor_multivar(self,
#                          filespath,
#                          vars,
#                          outpath = '/home/Earth/jpena/Suso/Coding/Python/',
#                         stats_fileout = 'stats.csv',
#                         distrib_fileout = 'distributions.csv',
#                         bins = 10, OutIn = True, Ni_files = 12):

#         cols_name = ['Minimum', 'Maximum',
#                     'Mean', 'Median',
#                     'Skewness', 'Kurtosis',
#                     'Std', 'Std_menos', 'Std_mas', 
#                     'Mad', 'Mad_menos', 'Mad_mas',
#                     'MC', 'Q1', 'Q3', 'IQD',
#                     'k_sd_min', 'k_sd_max',
#                     'k_sd_menos_min', 'k_sd_mas_max',
#                     'k_mad_min', 'k_mad_max',
#                     'k_mad_menos_min', 'k_mad_mas_max']

#         stats_vars = pd.DataFrame(columns = cols_name)
#         distribution_vars = pd.DataFrame() # = pd.Series()

#         for var_i in vars:

#             i_ = vars[0].find('_')
#             if i_>0:
#                 var_ii = var_i[:i_]
#             else:
#                 var_ii = var_i
#             filespath_var_i = filespath + var_i + '/'

#             print('Reading {}{}'.format(filespath_var_i,var_ii))

#             #stats_var_i, distribution_record, xbin = avg_online_stats_multi(filespath_var_i, var_ii, bins = bins, OutIn = OutIn, Ni_files = Ni_files)
#             stats_var_i, distribution_record, xbin = avg_online_stats(filespath_var_i, var_ii, bins = bins, OutIn = OutIn)


#             s_stats_var_i = pd.Series( stats_var_i, name = var_ii, index = cols_name)
#             stats_vars = stats_vars.append(s_stats_var_i)

#             s_distrib_var_i = pd.Series( distribution_record, name = var_ii) # = pd.Series( distribution_record, index = [var_i]*bins, xbin)
#             distribution_vars = distribution_vars.append(s_distrib_var_i) # = distribution_vars.append(s_distrib_var_i)
#             s_distrib_var_i_xbin = pd.Series( xbin, name = var_ii + '_xbin') # = pd.Series( distribution_record, index = [var_i]*bins, xbin)
#             distribution_vars = distribution_vars.append(s_distrib_var_i_xbin) # = distribution_vars.append(s_distrib_var_i)

#             print(' ')
#             print(var_i + ' added.')
#             print(' ')


#         stats_vars.to_csv(outpath + stats_fileout)
#         distribution_vars.to_csv(outpath + distrib_fileout)


#         return stats_vars, distribution_vars
            

    
    
if __name__ == '__main__':
    
    sys.exit()
       